import contextlib
import glob
import logging
import ntpath
import os
import posixpath
import re

from pathlib import Path, PosixPath, WindowsPath
from typing import (
    AsyncIterator,
    Callable,
    Final,
    Iterator,
    MutableMapping,
    Optional,
    Pattern,
    Tuple,
    Union,
)
from urllib.parse import urlparse

import lavalink

from redbot.core.i18n import Translator
from redbot.core.utils import AsyncIter

_ = Translator("Audio", Path(__file__))

_RE_REMOVE_START: Final[Pattern] = re.compile(r"^(sc|list) ")
_RE_YOUTUBE_TIMESTAMP: Final[Pattern] = re.compile(r"[&|?]t=(\d+)s?")
_RE_YOUTUBE_INDEX: Final[Pattern] = re.compile(r"&index=(\d+)")
_RE_SPOTIFY_URL: Final[Pattern] = re.compile(r"(http[s]?://)?(open\.spotify\.com)/")
_RE_SPOTIFY_TIMESTAMP: Final[Pattern] = re.compile(r"#(\d+):(\d+)")
_RE_SOUNDCLOUD_TIMESTAMP: Final[Pattern] = re.compile(r"#t=(\d+):(\d+)s?")
_RE_TWITCH_TIMESTAMP: Final[Pattern] = re.compile(r"\?t=(\d+)h(\d+)m(\d+)s")
_PATH_SEPS: Final[Tuple[str, str]] = (posixpath.sep, ntpath.sep)

_FULLY_SUPPORTED_MUSIC_EXT: Final[Tuple[str, ...]] = (".mp3", ".flac", ".ogg")
_PARTIALLY_SUPPORTED_MUSIC_EXT: Tuple[str, ...] = (
    ".m3u",
    ".m4a",
    ".aac",
    ".ra",
    ".wav",
    ".opus",
    ".wma",
    ".ts",
    ".au",
    # These do not work
    # ".mid",
    # ".mka",
    # ".amr",
    # ".aiff",
    # ".ac3",
    # ".voc",
    # ".dsf",
)
_PARTIALLY_SUPPORTED_VIDEO_EXT: Tuple[str, ...] = (
    ".mp4",
    ".mov",
    ".flv",
    ".webm",
    ".mkv",
    ".wmv",
    ".3gp",
    ".m4v",
    ".mk3d",  # https://github.com/Devoxin/lavaplayer
    ".mka",  # https://github.com/Devoxin/lavaplayer
    ".mks",  # https://github.com/Devoxin/lavaplayer
    # These do not work
    # ".vob",
    # ".mts",
    # ".avi",
    # ".mpg",
    # ".mpeg",
    # ".swf",
)
_PARTIALLY_SUPPORTED_MUSIC_EXT += _PARTIALLY_SUPPORTED_VIDEO_EXT


log = logging.getLogger("red.cogs.Audio.audio_dataclasses")


class LocalPath:
    """Local tracks class.

    Used to handle system dir trees in a cross system manner. The only use of this class is for
    `localtracks`.
    """

    _all_music_ext = _FULLY_SUPPORTED_MUSIC_EXT + _PARTIALLY_SUPPORTED_MUSIC_EXT

    def __init__(self, path, localtrack_folder, **kwargs):
        self._localtrack_folder = localtrack_folder
        self._path = path
        if isinstance(path, (Path, WindowsPath, PosixPath, LocalPath)):
            path = str(path.absolute())
        elif path is not None:
            path = str(path)

        self.cwd = Path.cwd()
        _lt_folder = Path(self._localtrack_folder) if self._localtrack_folder else self.cwd
        _path = Path(path) if path else self.cwd
        if _lt_folder.parts[-1].lower() == "localtracks" and not kwargs.get("forced"):
            self.localtrack_folder = _lt_folder
        elif kwargs.get("forced"):
            if _path.parts[-1].lower() == "localtracks":
                self.localtrack_folder = _path
            else:
                self.localtrack_folder = _path / "localtracks"
        else:
            self.localtrack_folder = _lt_folder / "localtracks"

        try:
            _path = Path(path)
            _path.relative_to(self.localtrack_folder)
            self.path = _path
        except (ValueError, TypeError):
            for sep in _PATH_SEPS:
                if path and path.startswith(f"localtracks{sep}{sep}"):
                    path = path.replace(f"localtracks{sep}{sep}", "", 1)
                elif path and path.startswith(f"localtracks{sep}"):
                    path = path.replace(f"localtracks{sep}", "", 1)
            self.path = self.localtrack_folder.joinpath(path) if path else self.localtrack_folder

        try:
            if self.path.is_file():
                parent = self.path.parent
            else:
                parent = self.path
            self.parent = Path(parent)
        except OSError:
            self.parent = None

    @property
    def name(self):
        return str(self.path.name)

    @property
    def suffix(self):
        return str(self.path.suffix)

    def is_dir(self):
        try:
            return self.path.is_dir()
        except OSError:
            return False

    def exists(self):
        try:
            return self.path.exists()
        except OSError:
            return False

    def is_file(self):
        try:
            return self.path.is_file()
        except OSError:
            return False

    def absolute(self):
        try:
            return self.path.absolute()
        except OSError:
            return self._path

    @classmethod
    def joinpath(cls, localpath, *args):
        modified = cls(None, localpath)
        modified.path = modified.path.joinpath(*args)
        return modified

    def rglob(self, pattern, folder=False) -> Iterator[str]:
        if folder:
            return glob.iglob(f"{glob.escape(self.path)}{os.sep}**{os.sep}", recursive=True)
        else:
            return glob.iglob(
                f"{glob.escape(self.path)}{os.sep}**{os.sep}*{pattern}", recursive=True
            )

    def glob(self, pattern, folder=False) -> Iterator[str]:
        if folder:
            return glob.iglob(f"{glob.escape(self.path)}{os.sep}*{os.sep}", recursive=False)
        else:
            return glob.iglob(f"{glob.escape(self.path)}{os.sep}*{pattern}", recursive=False)

    async def _multiglob(self, pattern: str, folder: bool, method: Callable):
        async for rp in AsyncIter(method(pattern)):
            rp_local = LocalPath(rp, self._localtrack_folder)
            if (
                (folder and rp_local.is_dir() and rp_local.exists())
                or (not folder and rp_local.suffix in self._all_music_ext and rp_local.is_file())
                and rp_local.exists()
            ):
                yield rp_local

    async def multiglob(self, *patterns, folder=False) -> AsyncIterator["LocalPath"]:
        async for p in AsyncIter(patterns):
            async for path in self._multiglob(p, folder, self.glob):
                yield path

    async def multirglob(self, *patterns, folder=False) -> AsyncIterator["LocalPath"]:
        async for p in AsyncIter(patterns):
            async for path in self._multiglob(p, folder, self.rglob):
                yield path

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return str(self)

    def to_string(self):
        try:
            return str(self.path.absolute())
        except OSError:
            return str(self._path)

    def to_string_user(self, arg: str = None):
        string = str(self.absolute()).replace(
            (str(self.localtrack_folder.absolute()) + os.sep) if arg is None else arg, ""
        )
        chunked = False
        while len(string) > 145 and os.sep in string:
            string = string.split(os.sep, 1)[-1]
            chunked = True

        if chunked:
            string = f"...{os.sep}{string}"
        return string

    async def tracks_in_tree(self):
        tracks = []
        async for track in self.multirglob(*[f"{ext}" for ext in self._all_music_ext]):
            with contextlib.suppress(ValueError):
                if track.path.parent != self.localtrack_folder and track.path.relative_to(
                    self.path
                ):
                    tracks.append(Query.process_input(track, self._localtrack_folder))
        return sorted(tracks, key=lambda x: x.to_string_user().lower())

    async def subfolders_in_tree(self):
        return_folders = []
        async for f in self.multirglob("", folder=True):
            with contextlib.suppress(ValueError):
                if (
                    f not in return_folders
                    and f.is_dir()
                    and f.path != self.localtrack_folder
                    and f.path.relative_to(self.path)
                ):
                    return_folders.append(f)
        return sorted(return_folders, key=lambda x: x.to_string_user().lower())

    async def tracks_in_folder(self):
        tracks = []
        async for track in self.multiglob(*[f"{ext}" for ext in self._all_music_ext]):
            with contextlib.suppress(ValueError):
                if track.path.parent != self.localtrack_folder and track.path.relative_to(
                    self.path
                ):
                    tracks.append(Query.process_input(track, self._localtrack_folder))
        return sorted(tracks, key=lambda x: x.to_string_user().lower())

    async def subfolders(self):
        return_folders = []
        async for f in self.multiglob("", folder=True):
            with contextlib.suppress(ValueError):
                if (
                    f not in return_folders
                    and f.path != self.localtrack_folder
                    and f.path.relative_to(self.path)
                ):
                    return_folders.append(f)
        return sorted(return_folders, key=lambda x: x.to_string_user().lower())

    def __eq__(self, other):
        if isinstance(other, LocalPath):
            return self.path._cparts == other.path._cparts
        elif isinstance(other, Path):
            return self.path._cparts == other._cpart
        return NotImplemented

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(tuple(self.path._cparts))
            return self._hash

    def __lt__(self, other):
        if isinstance(other, LocalPath):
            return self.path._cparts < other.path._cparts
        elif isinstance(other, Path):
            return self.path._cparts < other._cpart
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, LocalPath):
            return self.path._cparts <= other.path._cparts
        elif isinstance(other, Path):
            return self.path._cparts <= other._cpart
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, LocalPath):
            return self.path._cparts > other.path._cparts
        elif isinstance(other, Path):
            return self.path._cparts > other._cpart
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, LocalPath):
            return self.path._cparts >= other.path._cparts
        elif isinstance(other, Path):
            return self.path._cparts >= other._cpart
        return NotImplemented


class Query:
    """Query data class.

    Use: Query.process_input(query, localtrack_folder) to generate the Query object.
    """

    def __init__(self, query: Union[LocalPath, str], local_folder_current_path: Path, **kwargs):
        query = kwargs.get("queryforced", query)
        self._raw: Union[LocalPath, str] = query
        self._local_folder_current_path = local_folder_current_path
        _localtrack: LocalPath = LocalPath(query, local_folder_current_path)

        self.valid: bool = query != "InvalidQueryPlaceHolderName"
        self.is_local: bool = kwargs.get("local", False)
        self.is_spotify: bool = kwargs.get("spotify", False)
        self.is_youtube: bool = kwargs.get("youtube", False)
        self.is_soundcloud: bool = kwargs.get("soundcloud", False)
        self.is_bandcamp: bool = kwargs.get("bandcamp", False)
        self.is_vimeo: bool = kwargs.get("vimeo", False)
        self.is_mixer: bool = kwargs.get("mixer", False)
        self.is_twitch: bool = kwargs.get("twitch", False)
        self.is_other: bool = kwargs.get("other", False)
        self.is_pornhub: bool = kwargs.get("pornhub", False)
        self.is_playlist: bool = kwargs.get("playlist", False)
        self.is_album: bool = kwargs.get("album", False)
        self.is_search: bool = kwargs.get("search", False)
        self.is_stream: bool = kwargs.get("stream", False)
        self.single_track: bool = kwargs.get("single", False)
        self.id: Optional[str] = kwargs.get("id", None)
        self.invoked_from: Optional[str] = kwargs.get("invoked_from", None)
        self.local_name: Optional[str] = kwargs.get("name", None)
        self.search_subfolders: bool = kwargs.get("search_subfolders", False)
        self.spotify_uri: Optional[str] = kwargs.get("uri", None)
        self.uri: Optional[str] = kwargs.get("url", None)
        self.is_url: bool = kwargs.get("is_url", False)

        self.start_time: int = kwargs.get("start_time", 0)
        self.track_index: Optional[int] = kwargs.get("track_index", None)
        if self.invoked_from == "sc search":
            self.is_youtube = False
            self.is_soundcloud = True

        if (_localtrack.is_file() or _localtrack.is_dir()) and _localtrack.exists():
            self.local_track_path: Optional[LocalPath] = _localtrack
            self.track: str = str(_localtrack.absolute())
            self.is_local: bool = True
            self.uri = self.track
        else:
            self.local_track_path: Optional[LocalPath] = None
            self.track: str = str(query)

        self.lavalink_query: str = self._get_query()

        if self.is_playlist or self.is_album:
            self.single_track = False
        self._hash = hash(
            (
                self.valid,
                self.is_local,
                self.is_spotify,
                self.is_youtube,
                self.is_soundcloud,
                self.is_bandcamp,
                self.is_vimeo,
                self.is_mixer,
                self.is_twitch,
                self.is_other,
                self.is_playlist,
                self.is_album,
                self.is_search,
                self.is_stream,
                self.single_track,
                self.id,
                self.spotify_uri,
                self.start_time,
                self.track_index,
                self.uri,
            )
        )

    def __str__(self):
        return str(self.lavalink_query)

    @classmethod
    def process_input(
        cls,
        query: Union[LocalPath, lavalink.Track, "Query", str],
        _local_folder_current_path: Path,
        **kwargs,
    ) -> "Query":
        """Process the input query into its type.

        Parameters
        ----------
        query : Union[Query, LocalPath, lavalink.Track, str]
            The query string or LocalPath object.
        _local_folder_current_path: Path
            The Current Local Track folder
        Returns
        -------
        Query
            Returns a parsed Query object.
        """
        if not query:
            query = "InvalidQueryPlaceHolderName"
        possible_values = {}

        if isinstance(query, str):
            query = query.strip("<>")
            while "ytsearch:" in query:
                query = query.replace("ytsearch:", "")
            while "scsearch:" in query:
                query = query.replace("scsearch:", "")

        elif isinstance(query, Query):
            for key, val in kwargs.items():
                setattr(query, key, val)
            return query
        elif isinstance(query, lavalink.Track):
            possible_values["stream"] = query.is_stream
            query = query.uri

        possible_values.update(dict(**kwargs))
        possible_values.update(cls._parse(query, _local_folder_current_path, **kwargs))
        return cls(query, _local_folder_current_path, **possible_values)

    @staticmethod
    def _parse(track, _local_folder_current_path: Path, **kwargs) -> MutableMapping:
        """Parse a track into all the relevant metadata."""
        returning: MutableMapping = {}
        if (
            type(track) == type(LocalPath)
            and (track.is_file() or track.is_dir())
            and track.exists()
        ):
            returning["local"] = True
            returning["name"] = track.name
            if track.is_file():
                returning["single"] = True
            elif track.is_dir():
                returning["album"] = True
        else:
            track = str(track)
            if track.startswith("spotify:"):
                returning["spotify"] = True
                if ":playlist:" in track:
                    returning["playlist"] = True
                elif ":album:" in track:
                    returning["album"] = True
                elif ":track:" in track:
                    returning["single"] = True
                _id = track.split(":", 2)[-1]
                _id = _id.split("?")[0]
                returning["id"] = _id
                if "#" in _id:
                    match = re.search(_RE_SPOTIFY_TIMESTAMP, track)
                    if match:
                        returning["start_time"] = (int(match.group(1)) * 60) + int(match.group(2))
                returning["uri"] = track
                return returning
            if track.startswith("sc ") or track.startswith("list "):
                if track.startswith("sc "):
                    returning["invoked_from"] = "sc search"
                    returning["soundcloud"] = True
                elif track.startswith("list "):
                    returning["invoked_from"] = "search list"
                track = _RE_REMOVE_START.sub("", track, 1)
                returning["queryforced"] = track

            _localtrack = LocalPath(track, _local_folder_current_path)
            if _localtrack.exists():
                if _localtrack.is_file():
                    returning["local"] = True
                    returning["single"] = True
                    returning["name"] = _localtrack.name
                    return returning
                elif _localtrack.is_dir():
                    returning["album"] = True
                    returning["local"] = True
                    returning["name"] = _localtrack.name
                    return returning
            try:
                query_url = urlparse(track)
                if all([query_url.scheme, query_url.netloc, query_url.path]):
                    returning["url"] = track
                    returning["is_url"] = True
                    url_domain = ".".join(query_url.netloc.split(".")[-2:])
                    if not query_url.netloc:
                        url_domain = ".".join(query_url.path.split("/")[0].split(".")[-2:])
                    if url_domain in ["youtube.com", "youtu.be"]:
                        returning["youtube"] = True
                        _has_index = "&index=" in track
                        if "&t=" in track or "?t=" in track:
                            match = re.search(_RE_YOUTUBE_TIMESTAMP, track)
                            if match:
                                returning["start_time"] = int(match.group(1))
                        if _has_index:
                            match = re.search(_RE_YOUTUBE_INDEX, track)
                            if match:
                                returning["track_index"] = int(match.group(1)) - 1
                        if all(k in track for k in ["&list=", "watch?"]):
                            returning["track_index"] = 0
                            returning["playlist"] = True
                            returning["single"] = False
                        elif all(x in track for x in ["playlist?"]):
                            returning["playlist"] = not _has_index
                            returning["single"] = _has_index
                        elif any(k in track for k in ["list="]):
                            returning["track_index"] = 0
                            returning["playlist"] = True
                            returning["single"] = False
                        else:
                            returning["single"] = True
                    elif url_domain == "spotify.com":
                        returning["spotify"] = True
                        if "/playlist/" in track:
                            returning["playlist"] = True
                        elif "/album/" in track:
                            returning["album"] = True
                        elif "/track/" in track:
                            returning["single"] = True
                        val = re.sub(_RE_SPOTIFY_URL, "", track).replace("/", ":")
                        if "user:" in val:
                            val = val.split(":", 2)[-1]
                        _id = val.split(":", 1)[-1]
                        _id = _id.split("?")[0]

                        if "#" in _id:
                            _id = _id.split("#")[0]
                            match = re.search(_RE_SPOTIFY_TIMESTAMP, track)
                            if match:
                                returning["start_time"] = (int(match.group(1)) * 60) + int(
                                    match.group(2)
                                )

                        returning["id"] = _id
                        returning["uri"] = f"spotify:{val}"
                    elif url_domain == "soundcloud.com":
                        returning["soundcloud"] = True
                        if "#t=" in track:
                            match = re.search(_RE_SOUNDCLOUD_TIMESTAMP, track)
                            if match:
                                returning["start_time"] = (int(match.group(1)) * 60) + int(
                                    match.group(2)
                                )
                        if "/sets/" in track:
                            if "?in=" in track:
                                returning["single"] = True
                            else:
                                returning["playlist"] = True
                        else:
                            returning["single"] = True
                    elif url_domain == "bandcamp.com":
                        returning["bandcamp"] = True
                        if "/album/" in track:
                            returning["album"] = True
                        else:
                            returning["single"] = True
                    elif url_domain == "vimeo.com":
                        returning["vimeo"] = True
                    elif url_domain == "twitch.tv":
                        returning["twitch"] = True
                        if "?t=" in track:
                            match = re.search(_RE_TWITCH_TIMESTAMP, track)
                            if match:
                                returning["start_time"] = (
                                    (int(match.group(1)) * 60 * 60)
                                    + (int(match.group(2)) * 60)
                                    + int(match.group(3))
                                )

                        if not any(x in track for x in ["/clip/", "/videos/"]):
                            returning["stream"] = True
                    else:
                        returning["other"] = True
                        returning["single"] = True
                else:
                    if kwargs.get("soundcloud", False):
                        returning["soundcloud"] = True
                    else:
                        returning["youtube"] = True
                    returning["search"] = True
                    returning["single"] = True
            except Exception:
                returning["search"] = True
                returning["youtube"] = True
                returning["single"] = True
        return returning

    def _get_query(self):
        if self.is_local:
            return self.local_track_path.to_string()
        elif self.is_spotify:
            return self.spotify_uri
        elif self.is_search and self.is_youtube:
            return f"ytsearch:{self.track}"
        elif self.is_search and self.is_soundcloud:
            return f"scsearch:{self.track}"
        return self.track

    def to_string_user(self):
        if self.is_local:
            return str(self.local_track_path.to_string_user())
        return str(self._raw)

    @property
    def suffix(self):
        if self.is_local:
            return self.local_track_path.suffix
        return None

    def __eq__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return self.to_string_user() == other.to_string_user()

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            self._hash = hash(
                (
                    self.valid,
                    self.is_local,
                    self.is_spotify,
                    self.is_youtube,
                    self.is_soundcloud,
                    self.is_bandcamp,
                    self.is_vimeo,
                    self.is_mixer,
                    self.is_twitch,
                    self.is_other,
                    self.is_playlist,
                    self.is_album,
                    self.is_search,
                    self.is_stream,
                    self.single_track,
                    self.id,
                    self.spotify_uri,
                    self.start_time,
                    self.track_index,
                    self.uri,
                )
            )
            return self._hash

    def __lt__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return self.to_string_user() < other.to_string_user()

    def __le__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return self.to_string_user() <= other.to_string_user()

    def __gt__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return self.to_string_user() > other.to_string_user()

    def __ge__(self, other):
        if not isinstance(other, Query):
            return NotImplemented
        return self.to_string_user() >= other.to_string_user()



import asyncio
from datetime import datetime, timezone

from typing import Optional, Union

import discord

from redbot.core import checks, commands, modlog
from redbot.core.bot import Red
from redbot.core.i18n import Translator, cog_i18n
from redbot.core.utils.chat_formatting import box, pagify
from redbot.core.utils.menus import DEFAULT_CONTROLS, menu
from redbot.core.utils.predicates import MessagePredicate

_ = Translator("ModLog", __file__)


@cog_i18n(_)
class ModLog(commands.Cog):
    """Manage log channels for moderation actions."""

    def __init__(self, bot: Red):
        super().__init__()
        self.bot = bot

    async def red_delete_data_for_user(self, **kwargs):
        """ Nothing to delete """
        return

    @commands.group()
    @checks.guildowner_or_permissions(administrator=True)
    async def modlogset(self, ctx: commands.Context):
        """Manage modlog settings."""
        pass

    @checks.is_owner()
    @modlogset.command(hidden=True, name="fixcasetypes")
    async def reapply_audittype_migration(self, ctx: commands.Context):
        """Command to fix misbehaving casetypes."""
        await modlog.handle_auditype_key()
        await ctx.tick()

    @modlogset.command(aliases=["channel"])
    @commands.guild_only()
    async def modlog(self, ctx: commands.Context, channel: discord.TextChannel = None):
        """Set a channel as the modlog.

        Omit `[channel]` to disable the modlog.
        """
        guild = ctx.guild
        if channel:
            if channel.permissions_for(guild.me).send_messages:
                await modlog.set_modlog_channel(guild, channel)
                await ctx.send(
                    _("Mod events will be sent to {channel}.").format(channel=channel.mention)
                )
            else:
                await ctx.send(
                    _("I do not have permissions to send messages in {channel}!").format(
                        channel=channel.mention
                    )
                )
        else:
            try:
                await modlog.get_modlog_channel(guild)
            except RuntimeError:
                await ctx.send(_("Mod log is already disabled."))
            else:
                await modlog.set_modlog_channel(guild, None)
                await ctx.send(_("Mod log deactivated."))

    @modlogset.command(name="cases")
    @commands.guild_only()
    async def set_cases(self, ctx: commands.Context, action: str = None):
        """Enable or disable case creation for a mod action."""
        guild = ctx.guild

        if action is None:  # No args given
            casetypes = await modlog.get_all_casetypes(guild)
            await ctx.send_help()
            lines = []
            for ct in casetypes:
                enabled = _("enabled") if await ct.is_enabled() else _("disabled")
                lines.append(f"{ct.name} : {enabled}")

            await ctx.send(_("Current settings:\n") + box("\n".join(lines)))
            return

        casetype = await modlog.get_casetype(action, guild)
        if not casetype:
            await ctx.send(_("That action is not registered."))
        else:
            enabled = await casetype.is_enabled()
            await casetype.set_enabled(not enabled)
            await ctx.send(
                _("Case creation for {action_name} actions is now {enabled}.").format(
                    action_name=action, enabled=_("enabled") if not enabled else _("disabled")
                )
            )

    @modlogset.command()
    @commands.guild_only()
    async def resetcases(self, ctx: commands.Context):
        """Reset all modlog cases in this server."""
        guild = ctx.guild
        await ctx.send(
            _("Are you sure you would like to reset all modlog cases in this server?")
            + " (yes/no)"
        )
        try:
            pred = MessagePredicate.yes_or_no(ctx, user=ctx.author)
            msg = await ctx.bot.wait_for("message", check=pred, timeout=30)
        except asyncio.TimeoutError:
            await ctx.send(_("You took too long to respond."))
            return
        if pred.result:
            await modlog.reset_cases(guild)
            await ctx.send(_("Cases have been reset."))
        else:
            await ctx.send(_("No changes have been made."))

    @commands.command()
    @commands.guild_only()
    async def case(self, ctx: commands.Context, number: int):
        """Show the specified case."""
        try:
            case = await modlog.get_case(number, ctx.guild, self.bot)
        except RuntimeError:
            await ctx.send(_("That case does not exist for that server."))
            return
        else:
            if await ctx.embed_requested():
                await ctx.send(embed=await case.message_content(embed=True))
            else:
                message = _("{case}\n**Timestamp:** {timestamp}").format(
                    case=await case.message_content(embed=False),
                    timestamp=datetime.utcfromtimestamp(case.created_at).strftime(
                        "%Y-%m-%d %H:%M:%S UTC"
                    ),
                )
                await ctx.send(message)

    @commands.command()
    @commands.guild_only()
    async def casesfor(self, ctx: commands.Context, *, member: Union[discord.Member, int]):
        """Display cases for the specified member."""
        async with ctx.typing():
            try:
                if isinstance(member, int):
                    cases = await modlog.get_cases_for_member(
                        bot=ctx.bot, guild=ctx.guild, member_id=member
                    )
                else:
                    cases = await modlog.get_cases_for_member(
                        bot=ctx.bot, guild=ctx.guild, member=member
                    )
            except discord.NotFound:
                return await ctx.send(_("That user does not exist."))
            except discord.HTTPException:
                return await ctx.send(
                    _("Something unexpected went wrong while fetching that user by ID.")
                )

            if not cases:
                return await ctx.send(_("That user does not have any cases."))

            embed_requested = await ctx.embed_requested()
            if embed_requested:
                rendered_cases = [await case.message_content(embed=True) for case in cases]
            elif not embed_requested:
                rendered_cases = []
                for case in cases:
                    message = _("{case}\n**Timestamp:** {timestamp}").format(
                        case=await case.message_content(embed=False),
                        timestamp=datetime.utcfromtimestamp(case.created_at).strftime(
                            "%Y-%m-%d %H:%M:%S UTC"
                        ),
                    )
                    rendered_cases.append(message)

        await menu(ctx, rendered_cases, DEFAULT_CONTROLS)

    @commands.command()
    @commands.guild_only()
    async def listcases(self, ctx: commands.Context, *, member: Union[discord.Member, int]):
        """List cases for the specified member."""
        async with ctx.typing():
            try:
                if isinstance(member, int):
                    cases = await modlog.get_cases_for_member(
                        bot=ctx.bot, guild=ctx.guild, member_id=member
                    )
                else:
                    cases = await modlog.get_cases_for_member(
                        bot=ctx.bot, guild=ctx.guild, member=member
                    )
            except discord.NotFound:
                return await ctx.send(_("That user does not exist."))
            except discord.HTTPException:
                return await ctx.send(
                    _("Something unexpected went wrong while fetching that user by ID.")
                )
            if not cases:
                return await ctx.send(_("That user does not have any cases."))

            rendered_cases = []
            message = ""
            for case in cases:
                message += _("{case}\n**Timestamp:** {timestamp}\n\n").format(
                    case=await case.message_content(embed=False),
                    timestamp=datetime.utcfromtimestamp(case.created_at).strftime(
                        "%Y-%m-%d %H:%M:%S UTC"
                    ),
                )
            for page in pagify(message, ["\n\n", "\n"], priority=True):
                rendered_cases.append(page)
        await menu(ctx, rendered_cases, DEFAULT_CONTROLS)

    @commands.command()
    @commands.guild_only()
    async def reason(self, ctx: commands.Context, case: Optional[int], *, reason: str):
        """Specify a reason for a modlog case.

        Please note that you can only edit cases you are
        the owner of unless you are a mod, admin or server owner.

        If no case number is specified, the latest case will be used.
        """
        author = ctx.author
        guild = ctx.guild
        if case is None:
            # get the latest case
            case_obj = await modlog.get_latest_case(guild, self.bot)
            if case_obj is None:
                await ctx.send(_("There are no modlog cases in this server."))
                return
        else:
            try:
                case_obj = await modlog.get_case(case, guild, self.bot)
            except RuntimeError:
                await ctx.send(_("That case does not exist!"))
                return

        is_guild_owner = author == guild.owner
        is_case_author = author == case_obj.moderator
        author_is_mod = await ctx.bot.is_mod(author)
        if not (is_guild_owner or is_case_author or author_is_mod):
            await ctx.send(_("You are not authorized to modify that case!"))
            return
        to_modify = {"reason": reason}
        if case_obj.moderator != author:
            to_modify["amended_by"] = author
        to_modify["modified_at"] = ctx.message.created_at.replace(tzinfo=timezone.utc).timestamp()
        await case_obj.edit(to_modify)
        await ctx.send(
            _("Reason for case #{num} has been updated.").format(num=case_obj.case_number)
        )



import discord
from redbot.core import commands
from redbot.core.i18n import Translator
from .installable import InstalledModule

_ = Translator("Koala", __file__)


class InstalledCog(InstalledModule):
    @classmethod
    async def convert(cls, ctx: commands.Context, arg: str) -> InstalledModule:
        downloader = ctx.bot.get_cog("Downloader")
        if downloader is None:
            raise commands.CommandError(_("No Downloader cog found."))

        cog = discord.utils.get(await downloader.installed_cogs(), name=arg)
        if cog is None:
            raise commands.BadArgument(
                _("Cog `{cog_name}` is not installed.").format(cog_name=arg)
            )

        return cog



import asyncio
import logging
from pathlib import Path

import lavalink

from redbot.core import data_manager
from redbot.core.i18n import Translator
from ...errors import LavalinkDownloadFailed
from ...manager import ServerManager
from ..abc import MixinMeta
from ..cog_utils import CompositeMetaClass

log = logging.getLogger("red.cogs.Audio.cog.Tasks.lavalink")
_ = Translator("Audio", Path(__file__))


class LavalinkTasks(MixinMeta, metaclass=CompositeMetaClass):
    def lavalink_restart_connect(self) -> None:
        lavalink.unregister_event_listener(self.lavalink_event_handler)
        lavalink.unregister_update_listener(self.lavalink_update_handler)
        if self.lavalink_connect_task:
            self.lavalink_connect_task.cancel()
        if self._restore_task:
            self._restore_task.cancel()

        self._restore_task = None
        lavalink.register_event_listener(self.lavalink_event_handler)
        lavalink.register_update_listener(self.lavalink_update_handler)
        self.lavalink_connect_task = self.bot.loop.create_task(self.lavalink_attempt_connect())

    async def lavalink_attempt_connect(self, timeout: int = 50) -> None:
        self.lavalink_connection_aborted = False
        max_retries = 5
        retry_count = 0
        while retry_count < max_retries:
            configs = await self.config.all()
            external = configs["use_external_lavalink"]
            java_exec = configs["java_exc_path"]
            if external is False:
                settings = self._default_lavalink_settings
                host = settings["host"]
                password = settings["password"]
                ws_port = settings["ws_port"]
                if self.player_manager is not None:
                    await self.player_manager.shutdown()
                self.player_manager = ServerManager()
                try:
                    await self.player_manager.start(java_exec)
                except LavalinkDownloadFailed as exc:
                    await asyncio.sleep(1)
                    if exc.should_retry:
                        log.exception(
                            "Exception whilst starting internal Lavalink server, retrying...",
                            exc_info=exc,
                        )
                        retry_count += 1
                        continue
                    else:
                        log.exception(
                            "Fatal exception whilst starting internal Lavalink server, "
                            "aborting...",
                            exc_info=exc,
                        )
                        self.lavalink_connection_aborted = True
                        raise
                except asyncio.CancelledError:
                    log.exception("Invalid machine architecture, cannot run Lavalink.")
                    raise
                except Exception as exc:
                    log.exception(
                        "Unhandled exception whilst starting internal Lavalink server, "
                        "aborting...",
                        exc_info=exc,
                    )
                    self.lavalink_connection_aborted = True
                    raise
                else:
                    break
            else:
                host = configs["host"]
                password = configs["password"]
                ws_port = configs["ws_port"]
                break
        else:
            log.critical(
                "Setting up the Lavalink server failed after multiple attempts. "
                "See above tracebacks for details."
            )
            self.lavalink_connection_aborted = True
            return

        retry_count = 0
        while retry_count < max_retries:
            if lavalink.node._nodes:
                await lavalink.node.disconnect()
            try:
                await lavalink.initialize(
                    bot=self.bot,
                    host=host,
                    password=password,
                    ws_port=ws_port,
                    timeout=timeout,
                    resume_key=f"Red-Core-Audio-{self.bot.user.id}-{data_manager.instance_name}",
                )
            except asyncio.TimeoutError:
                log.error("Connecting to Lavalink server timed out, retrying...")
                if external is False and self.player_manager is not None:
                    await self.player_manager.shutdown()
                retry_count += 1
                await asyncio.sleep(1)  # prevent busylooping
            except Exception as exc:
                log.exception(
                    "Unhandled exception whilst connecting to Lavalink, aborting...", exc_info=exc
                )
                self.lavalink_connection_aborted = True
                raise
            else:
                break
        else:
            self.lavalink_connection_aborted = True
            log.critical(
                "Connecting to the Lavalink server failed after multiple attempts. "
                "See above tracebacks for details."
            )
            return
        if external:
            await asyncio.sleep(5)
        self._restore_task = asyncio.create_task(self.restore_players())



import pytest

from redbot.cogs.permissions import Permissions
from redbot.core import Config


@pytest.fixture()
def permissions(config, monkeypatch, red):
    with monkeypatch.context() as m:
        m.setattr(Config, "get_conf", lambda *args, **kwargs: config)
        return Permissions(red)



from typing import Final

# TODO: https://github.com/Cog-Creators/Red-DiscordBot/pull/3195#issuecomment-567821701
# Thanks a lot Sinbad!

__all__ = [
    # PRAGMA Statements
    "PRAGMA_SET_temp_store",
    "PRAGMA_SET_journal_mode",
    "PRAGMA_SET_read_uncommitted",
    "PRAGMA_FETCH_user_version",
    "PRAGMA_SET_user_version",
    # Data Deletion statement
    "HANDLE_DISCORD_DATA_DELETION_QUERY",
    # Playlist table statements
    "PLAYLIST_CREATE_TABLE",
    "PLAYLIST_DELETE",
    "PLAYLIST_DELETE_SCOPE",
    "PLAYLIST_DELETE_SCHEDULED",
    "PLAYLIST_FETCH_ALL",
    "PLAYLIST_FETCH_ALL_WITH_FILTER",
    "PLAYLIST_FETCH_ALL_CONVERTER",
    "PLAYLIST_FETCH",
    "PLAYLIST_UPSERT",
    "PLAYLIST_CREATE_INDEX",
    # YouTube table statements
    "YOUTUBE_DROP_TABLE",
    "YOUTUBE_CREATE_TABLE",
    "YOUTUBE_CREATE_INDEX",
    "YOUTUBE_UPSERT",
    "YOUTUBE_UPDATE",
    "YOUTUBE_QUERY",
    "YOUTUBE_QUERY_ALL",
    "YOUTUBE_DELETE_OLD_ENTRIES",
    "YOUTUBE_QUERY_LAST_FETCHED_RANDOM",
    # Spotify table statements
    "SPOTIFY_DROP_TABLE",
    "SPOTIFY_CREATE_INDEX",
    "SPOTIFY_CREATE_TABLE",
    "SPOTIFY_UPSERT",
    "SPOTIFY_QUERY",
    "SPOTIFY_QUERY_ALL",
    "SPOTIFY_UPDATE",
    "SPOTIFY_DELETE_OLD_ENTRIES",
    "SPOTIFY_QUERY_LAST_FETCHED_RANDOM",
    # Lavalink table statements
    "LAVALINK_DROP_TABLE",
    "LAVALINK_CREATE_TABLE",
    "LAVALINK_CREATE_INDEX",
    "LAVALINK_UPSERT",
    "LAVALINK_UPDATE",
    "LAVALINK_QUERY",
    "LAVALINK_QUERY_ALL",
    "LAVALINK_QUERY_LAST_FETCHED_RANDOM",
    "LAVALINK_DELETE_OLD_ENTRIES",
    "LAVALINK_FETCH_ALL_ENTRIES_GLOBAL",
    # Persisting Queue statements
    "PERSIST_QUEUE_DROP_TABLE",
    "PERSIST_QUEUE_CREATE_TABLE",
    "PERSIST_QUEUE_CREATE_INDEX",
    "PERSIST_QUEUE_PLAYED",
    "PERSIST_QUEUE_DELETE_SCHEDULED",
    "PERSIST_QUEUE_FETCH_ALL",
    "PERSIST_QUEUE_UPSERT",
    "PERSIST_QUEUE_BULK_PLAYED",
]

# PRAGMA Statements

PRAGMA_SET_temp_store: Final[
    str
] = """
PRAGMA temp_store = 2;
"""
PRAGMA_SET_journal_mode: Final[
    str
] = """
PRAGMA journal_mode = wal;
"""
PRAGMA_SET_read_uncommitted: Final[
    str
] = """
PRAGMA read_uncommitted = 1;
"""
PRAGMA_FETCH_user_version: Final[
    str
] = """
pragma user_version;
"""
PRAGMA_SET_user_version: Final[
    str
] = """
pragma user_version=3;
"""

# Data Deletion
# This is intentionally 2 seperate transactions due to concerns
# Draper had. This should prevent it from being a large issue,
# as this is no different than triggering a bulk deletion now.
HANDLE_DISCORD_DATA_DELETION_QUERY: Final[
    str
] = """
BEGIN TRANSACTION;

UPDATE playlists
SET deleted = true
WHERE scope_id = :user_id ;

UPDATE playlists
SET author_id = 0xde1
WHERE author_id = :user_id ;

COMMIT TRANSACTION;

BEGIN TRANSACTION;

DELETE FROM PLAYLISTS
WHERE deleted=true;

COMMIT TRANSACTION;
"""

# Playlist table statements
PLAYLIST_CREATE_TABLE: Final[
    str
] = """
CREATE TABLE IF NOT EXISTS playlists (
    scope_type INTEGER NOT NULL,
    playlist_id INTEGER NOT NULL,
    playlist_name TEXT NOT NULL,
    scope_id INTEGER NOT NULL,
    author_id INTEGER NOT NULL,
    deleted BOOLEAN DEFAULT false,
    playlist_url TEXT,
    tracks JSON,
    PRIMARY KEY (playlist_id, scope_id, scope_type)
);
"""
PLAYLIST_DELETE: Final[
    str
] = """
UPDATE playlists
    SET
        deleted = true
WHERE
    (
        scope_type = :scope_type
        AND playlist_id = :playlist_id
        AND scope_id = :scope_id
    )
;
"""
PLAYLIST_DELETE_SCOPE: Final[
    str
] = """
DELETE
FROM
    playlists
WHERE
    scope_type = :scope_type ;
"""
PLAYLIST_DELETE_SCHEDULED: Final[
    str
] = """
DELETE
FROM
    playlists
WHERE
    deleted = true;
"""
PLAYLIST_FETCH_ALL: Final[
    str
] = """
SELECT
    playlist_id,
    playlist_name,
    scope_id,
    author_id,
    playlist_url,
    tracks
FROM
    playlists
WHERE
    scope_type = :scope_type
    AND scope_id = :scope_id
    AND deleted = false
    ;
"""
PLAYLIST_FETCH_ALL_WITH_FILTER: Final[
    str
] = """
SELECT
    playlist_id,
    playlist_name,
    scope_id,
    author_id,
    playlist_url,
    tracks
FROM
    playlists
WHERE
    (
        scope_type = :scope_type
        AND scope_id = :scope_id
        AND author_id = :author_id
        AND deleted = false
    )
;
"""
PLAYLIST_FETCH_ALL_CONVERTER: Final[
    str
] = """
SELECT
    playlist_id,
    playlist_name,
    scope_id,
    author_id,
    playlist_url,
    tracks
FROM
    playlists
WHERE
    (
        scope_type = :scope_type
        AND
        (
        playlist_id = :playlist_id
        OR
        LOWER(playlist_name) LIKE "%" || COALESCE(LOWER(:playlist_name), "") || "%"
        )
        AND deleted = false
    )
;
"""
PLAYLIST_FETCH: Final[
    str
] = """
SELECT
    playlist_id,
    playlist_name,
    scope_id,
    author_id,
    playlist_url,
    tracks
FROM
    playlists
WHERE
    (
        scope_type = :scope_type
        AND playlist_id = :playlist_id
        AND scope_id = :scope_id
        AND deleted = false
    )
LIMIT 1;
"""
PLAYLIST_UPSERT: Final[
    str
] = """
INSERT INTO
    playlists ( scope_type, playlist_id, playlist_name, scope_id, author_id, playlist_url, tracks )
VALUES
    (
        :scope_type, :playlist_id, :playlist_name, :scope_id, :author_id, :playlist_url, :tracks
    )
    ON CONFLICT (scope_type, playlist_id, scope_id) DO
    UPDATE
    SET
        playlist_name = excluded.playlist_name,
        playlist_url = excluded.playlist_url,
        tracks = excluded.tracks;
"""
PLAYLIST_CREATE_INDEX: Final[
    str
] = """
CREATE INDEX IF NOT EXISTS name_index ON playlists (
scope_type, playlist_id, playlist_name, scope_id
);
"""

# YouTube table statements
YOUTUBE_DROP_TABLE: Final[
    str
] = """
DROP TABLE IF EXISTS youtube;
"""
YOUTUBE_CREATE_TABLE: Final[
    str
] = """
CREATE TABLE IF NOT EXISTS youtube(
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    track_info TEXT,
    youtube_url TEXT,
    last_updated INTEGER,
    last_fetched INTEGER
);
"""
YOUTUBE_CREATE_INDEX: Final[
    str
] = """
CREATE UNIQUE INDEX IF NOT EXISTS idx_youtube_url
ON youtube (track_info, youtube_url);
"""
YOUTUBE_UPSERT: Final[
    str
] = """INSERT INTO
youtube
  (
    track_info,
    youtube_url,
    last_updated,
    last_fetched
  )
VALUES
  (
   :track_info,
   :track_url,
   :last_updated,
   :last_fetched
  )
ON CONFLICT
  (
  track_info,
  youtube_url
  )
DO UPDATE
  SET
    track_info = excluded.track_info,
    last_updated = excluded.last_updated
"""
YOUTUBE_UPDATE: Final[
    str
] = """
UPDATE youtube
SET last_fetched=:last_fetched
WHERE track_info=:track;
"""
YOUTUBE_QUERY: Final[
    str
] = """
SELECT youtube_url, last_updated
FROM youtube
WHERE
    track_info=:track
    AND last_updated > :maxage
LIMIT 1;
"""
YOUTUBE_QUERY_ALL: Final[
    str
] = """
SELECT youtube_url, last_updated
FROM youtube
"""
YOUTUBE_DELETE_OLD_ENTRIES: Final[
    str
] = """
DELETE FROM youtube
WHERE
    last_updated < :maxage
    ;
"""
YOUTUBE_QUERY_LAST_FETCHED_RANDOM: Final[
    str
] = """
SELECT youtube_url, last_updated
FROM youtube
WHERE
    last_fetched > :day
    AND last_updated > :maxage
LIMIT 100
;
"""

# Spotify table statements
SPOTIFY_DROP_TABLE: Final[
    str
] = """
DROP TABLE IF EXISTS spotify;
"""
SPOTIFY_CREATE_TABLE: Final[
    str
] = """
CREATE TABLE IF NOT EXISTS spotify(
    id TEXT,
    type TEXT,
    uri TEXT,
    track_name TEXT,
    artist_name TEXT,
    song_url TEXT,
    track_info TEXT,
    last_updated INTEGER,
    last_fetched INTEGER
);
"""
SPOTIFY_CREATE_INDEX: Final[
    str
] = """
CREATE UNIQUE INDEX IF NOT EXISTS idx_spotify_uri
ON spotify (id, type, uri);
"""
SPOTIFY_UPSERT: Final[
    str
] = """INSERT INTO
spotify
  (
    id, type, uri, track_name, artist_name,
    song_url, track_info, last_updated, last_fetched
  )
VALUES
  (
    :id, :type, :uri, :track_name, :artist_name,
    :song_url, :track_info, :last_updated, :last_fetched
  )
ON CONFLICT
  (
    id,
    type,
    uri
  )
DO UPDATE
  SET
    track_name = excluded.track_name,
    artist_name = excluded.artist_name,
    song_url = excluded.song_url,
    track_info = excluded.track_info,
    last_updated = excluded.last_updated;
"""
SPOTIFY_UPDATE: Final[
    str
] = """
UPDATE spotify
SET last_fetched=:last_fetched
WHERE uri=:uri;
"""
SPOTIFY_QUERY: Final[
    str
] = """
SELECT track_info, last_updated
FROM spotify
WHERE
    uri=:uri
    AND last_updated > :maxage
LIMIT 1;
"""
SPOTIFY_QUERY_ALL: Final[
    str
] = """
SELECT track_info, last_updated
FROM spotify
"""
SPOTIFY_DELETE_OLD_ENTRIES: Final[
    str
] = """
DELETE FROM spotify
WHERE
    last_updated < :maxage
    ;
"""
SPOTIFY_QUERY_LAST_FETCHED_RANDOM: Final[
    str
] = """
SELECT track_info, last_updated
FROM spotify
WHERE
    last_fetched > :day
    AND last_updated > :maxage
LIMIT 100
;
"""

# Lavalink table statements
LAVALINK_DROP_TABLE: Final[
    str
] = """
DROP TABLE IF EXISTS lavalink ;
"""
LAVALINK_CREATE_TABLE: Final[
    str
] = """
CREATE TABLE IF NOT EXISTS lavalink(
    query TEXT,
    data JSON,
    last_updated INTEGER,
    last_fetched INTEGER

);
"""
LAVALINK_CREATE_INDEX: Final[
    str
] = """
CREATE UNIQUE INDEX IF NOT EXISTS idx_lavalink_query
ON lavalink (query);
"""
LAVALINK_UPSERT: Final[
    str
] = """INSERT INTO
lavalink
  (
    query,
    data,
    last_updated,
    last_fetched
  )
VALUES
  (
   :query,
   :data,
   :last_updated,
   :last_fetched
  )
ON CONFLICT
  (
    query
  )
DO UPDATE
  SET
    data = excluded.data,
    last_updated = excluded.last_updated;
"""
LAVALINK_UPDATE: Final[
    str
] = """
UPDATE lavalink
SET last_fetched=:last_fetched
WHERE query=:query;
"""
LAVALINK_QUERY: Final[
    str
] = """
SELECT data, last_updated
FROM lavalink
WHERE
    query=:query
    AND last_updated > :maxage
LIMIT 1;
"""
LAVALINK_QUERY_ALL: Final[
    str
] = """
SELECT data, last_updated
FROM lavalink
"""
LAVALINK_QUERY_LAST_FETCHED_RANDOM: Final[
    str
] = """
SELECT data, last_updated
FROM lavalink
WHERE
    last_fetched > :day
    AND last_updated > :maxage
LIMIT 100
;
"""
LAVALINK_DELETE_OLD_ENTRIES: Final[
    str
] = """
DELETE FROM lavalink
WHERE
    last_updated < :maxage
    ;
"""
LAVALINK_FETCH_ALL_ENTRIES_GLOBAL: Final[
    str
] = """
SELECT query, data 
FROM lavalink
"""

# Persisting Queue statements
PERSIST_QUEUE_DROP_TABLE: Final[
    str
] = """
DROP TABLE IF EXISTS persist_queue ;
"""
PERSIST_QUEUE_CREATE_TABLE: Final[
    str
] = """
CREATE TABLE IF NOT EXISTS persist_queue(
    guild_id INTEGER NOT NULL,
    room_id INTEGER NOT NULL,
    track JSON NOT NULL,
    played BOOLEAN DEFAULT false,
    track_id TEXT NOT NULL,
    time INTEGER NOT NULL,
    PRIMARY KEY (guild_id, room_id, track_id)
);
"""
PERSIST_QUEUE_CREATE_INDEX: Final[
    str
] = """
CREATE INDEX IF NOT EXISTS track_index ON persist_queue (guild_id, track_id);
"""
PERSIST_QUEUE_PLAYED: Final[
    str
] = """
UPDATE persist_queue
    SET
        played = true
WHERE
    (
        guild_id = :guild_id
        AND track_id = :track_id
    )
;
"""
PERSIST_QUEUE_BULK_PLAYED: Final[
    str
] = """
UPDATE persist_queue
    SET
        played = true
WHERE guild_id = :guild_id
;
"""
PERSIST_QUEUE_DELETE_SCHEDULED: Final[
    str
] = """
DELETE
FROM
    persist_queue
WHERE
    played = true;
"""
PERSIST_QUEUE_FETCH_ALL: Final[
    str
] = """
SELECT
    guild_id, room_id, track
FROM
    persist_queue
WHERE played = false
ORDER BY time ASC;
"""
PERSIST_QUEUE_UPSERT: Final[
    str
] = """
INSERT INTO
    persist_queue (guild_id, room_id, track, played, track_id, time)
VALUES
    (
        :guild_id, :room_id, :track, :played, :track_id, :time
    )
ON CONFLICT (guild_id, room_id, track_id) DO
UPDATE
    SET
        time = excluded.time
"""



from ..cog_utils import CompositeMetaClass
from .equalizer import EqualizerUtilities
from .formatting import FormattingUtilities
from .local_tracks import LocalTrackUtilities
from .miscellaneous import MiscellaneousUtilities
from .parsers import ParsingUtilities
from .player import PlayerUtilities
from .playlists import PlaylistUtilities
from .queue import QueueUtilities
from .validation import ValidationUtilities


class Utilities(
    EqualizerUtilities,
    FormattingUtilities,
    LocalTrackUtilities,
    MiscellaneousUtilities,
    PlayerUtilities,
    PlaylistUtilities,
    QueueUtilities,
    ValidationUtilities,
    ParsingUtilities,
    metaclass=CompositeMetaClass,
):
    """Class joining all utility subclasses"""



from pathlib import Path
import subprocess as sp

import pytest

from redbot.cogs.downloader.repo_manager import ProcessFormatter, Repo
from redbot.pytest.downloader import (
    GIT_VERSION,
    cloned_git_repo,
    git_repo,
    git_repo_with_remote,
    _session_git_repo,
)


@pytest.mark.asyncio
async def test_git_clone_nobranch(git_repo, tmp_path):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CLONE_NO_BRANCH,
            url=git_repo.folder_path,
            folder=tmp_path / "cloned_repo_test",
        )
    )
    assert p.returncode == 0


@pytest.mark.asyncio
async def test_git_clone_branch(git_repo, tmp_path):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CLONE,
            branch="master",
            url=git_repo.folder_path,
            folder=tmp_path / "cloned_repo_test",
        )
    )
    assert p.returncode == 0


@pytest.mark.asyncio
async def test_git_clone_non_existent_branch(git_repo, tmp_path):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CLONE,
            branch="non-existent-branch",
            url=git_repo.folder_path,
            folder=tmp_path / "cloned_repo_test",
        )
    )
    assert p.returncode == 128


@pytest.mark.asyncio
async def test_git_clone_notgit_repo(git_repo, tmp_path):
    notgit_repo = tmp_path / "test_clone_folder"
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CLONE, branch=None, url=notgit_repo, folder=tmp_path / "cloned_repo_test"
        )
    )
    assert p.returncode == 128


@pytest.mark.asyncio
async def test_git_current_branch_master(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(git_repo.GIT_CURRENT_BRANCH, path=git_repo.folder_path)
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "master"


@pytest.mark.asyncio
async def test_git_current_branch_detached(git_repo):
    await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECKOUT,
            path=git_repo.folder_path,
            rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
        )
    )
    p = await git_repo._run(
        ProcessFormatter().format(git_repo.GIT_CURRENT_BRANCH, path=git_repo.folder_path)
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == "fatal: ref HEAD is not a symbolic ref"


@pytest.mark.asyncio
async def test_git_current_commit_on_branch(git_repo):
    # HEAD on dont_add_commits (a0ccc2390883c85a361f5a90c72e1b07958939fa)
    # setup
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECKOUT, path=git_repo.folder_path, rev="dont_add_commits"
        )
    )
    assert p.returncode == 0

    p = await git_repo._run(
        ProcessFormatter().format(git_repo.GIT_CURRENT_COMMIT, path=git_repo.folder_path)
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "a0ccc2390883c85a361f5a90c72e1b07958939fa"


@pytest.mark.asyncio
async def test_git_current_commit_detached(git_repo):
    # detached HEAD state (c950fc05a540dd76b944719c2a3302da2e2f3090)
    await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECKOUT,
            path=git_repo.folder_path,
            rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
        )
    )
    p = await git_repo._run(
        ProcessFormatter().format(git_repo.GIT_CURRENT_COMMIT, path=git_repo.folder_path)
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "c950fc05a540dd76b944719c2a3302da2e2f3090"


@pytest.mark.asyncio
async def test_git_latest_commit(git_repo):
    # HEAD on dont_add_commits (a0ccc2390883c85a361f5a90c72e1b07958939fa)
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_LATEST_COMMIT, path=git_repo.folder_path, branch="dont_add_commits"
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "a0ccc2390883c85a361f5a90c72e1b07958939fa"


@pytest.mark.asyncio
async def test_git_hard_reset(cloned_git_repo, tmp_path):
    staged_file = cloned_git_repo.folder_path / "staged_file.txt"
    staged_file.touch()
    git_dirparams = ("git", "-C", str(cloned_git_repo.folder_path))
    sp.run((*git_dirparams, "add", "staged_file.txt"), check=True)
    assert staged_file.exists() is True
    p = await cloned_git_repo._run(
        ProcessFormatter().format(
            cloned_git_repo.GIT_HARD_RESET, path=cloned_git_repo.folder_path, branch="master"
        )
    )
    assert p.returncode == 0
    assert staged_file.exists() is False


@pytest.mark.asyncio
async def test_git_pull(git_repo_with_remote, tmp_path):
    # setup
    staged_file = Path(git_repo_with_remote.url) / "staged_file.txt"
    staged_file.touch()
    git_dirparams = ("git", "-C", git_repo_with_remote.url)
    sp.run((*git_dirparams, "add", "staged_file.txt"), check=True)
    sp.run(
        (*git_dirparams, "commit", "-m", "test commit", "--no-gpg-sign", "--no-verify"), check=True
    )
    assert not (git_repo_with_remote.folder_path / "staged_file.txt").exists()

    p = await git_repo_with_remote._run(
        ProcessFormatter().format(
            git_repo_with_remote.GIT_PULL, path=git_repo_with_remote.folder_path
        )
    )
    assert p.returncode == 0
    assert (git_repo_with_remote.folder_path / "staged_file.txt").exists()


@pytest.mark.asyncio
async def test_git_diff_file_status(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_DIFF_FILE_STATUS,
            path=git_repo.folder_path,
            old_rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
            new_rev="fb99eb7d2d5bed514efc98fe6686b368f8425745",
        )
    )
    assert p.returncode == 0
    stdout = p.stdout.strip(b"\t\n\x00 ").decode()
    assert stdout == (
        "A\x00added_file.txt\x00\t"
        "M\x00mycog/__init__.py\x00\t"
        "D\x00sample_file1.txt\x00\t"
        "D\x00sample_file2.txt\x00\t"
        "A\x00sample_file3.txt"
    )


# might need to add test for test_git_log, but it's unused method currently


@pytest.mark.asyncio
async def test_git_discover_remote_url(cloned_git_repo, tmp_path):
    p = await cloned_git_repo._run(
        ProcessFormatter().format(
            cloned_git_repo.GIT_DISCOVER_REMOTE_URL, path=cloned_git_repo.folder_path
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == cloned_git_repo.url


@pytest.mark.asyncio
async def test_git_checkout_detached_head(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECKOUT,
            path=git_repo.folder_path,
            rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
        )
    )
    assert p.returncode == 0


@pytest.mark.asyncio
async def test_git_checkout_branch(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECKOUT, path=git_repo.folder_path, rev="dont_add_commits"
        )
    )
    assert p.returncode == 0


@pytest.mark.asyncio
async def test_git_checkout_non_existent_branch(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECKOUT, path=git_repo.folder_path, rev="non-existent-branch"
        )
    )
    assert p.returncode == 1


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_branch_name(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="dont_add_commits"
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "a0ccc2390883c85a361f5a90c72e1b07958939fa"


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_full_hash(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1,
            path=git_repo.folder_path,
            rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "c950fc05a540dd76b944719c2a3302da2e2f3090"


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_short_hash(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="c950"
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "c950fc05a540dd76b944719c2a3302da2e2f3090"


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_too_short_hash(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="c95")
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == "fatal: Needed a single revision"


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_lightweight_tag(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="lightweight"
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "fb99eb7d2d5bed514efc98fe6686b368f8425745"


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_annotated_tag(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="annotated"
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == "a7120330cc179396914e0d6af80cfa282adc124b"


@pytest.mark.asyncio
async def test_git_get_full_sha1_from_invalid_ref(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="invalid"
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == "fatal: Needed a single revision"


@pytest.mark.skipif(
    GIT_VERSION < (2, 31), reason="This is test for output from Git 2.31 and newer."
)
@pytest.mark.asyncio
async def test_git_get_full_sha1_from_ambiguous_commits(git_repo):
    # 2 ambiguous refs:
    # branch ambiguous_1 - 95da0b576271cb5bee5f3e075074c03ee05fed05
    # branch ambiguous_2 - 95da0b57a416d9c8ce950554228d1fc195c30b43
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="95da0b57"
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == (
        "error: short object ID 95da0b57 is ambiguous\n"
        "hint: The candidates are:\n"
        "hint:   95da0b576 commit 2019-10-22 - Ambiguous commit 16955\n"
        "hint:   95da0b57a commit 2019-10-22 - Ambiguous commit 44414\n"
        "fatal: Needed a single revision"
    )


@pytest.mark.skipif(
    GIT_VERSION < (2, 31), reason="This is test for output from Git 2.31 and newer."
)
@pytest.mark.asyncio
async def test_git_get_full_sha1_from_ambiguous_tag_and_commit(git_repo):
    # 2 ambiguous refs:
    # branch ambiguous_with_tag - c6f0e5ec04d99bdf8c6c78ff20d66d286eecb3ea
    # tag ambiguous_tag_66387 - c6f0e5ec04d99bdf8c6c78ff20d66d286eecb3ea
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="c6f0"
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == (
        "error: short object ID c6f0 is ambiguous\n"
        "hint: The candidates are:\n"
        "hint:   c6f028f tag ambiguous_tag_66387\n"
        "hint:   c6f0e5e commit 2019-10-24 - Commit ambiguous with tag.\n"
        "fatal: Needed a single revision"
    )


@pytest.mark.skipif(
    GIT_VERSION >= (2, 31), reason="This is test for output from Git older than 2.31."
)
@pytest.mark.asyncio
async def test_git_get_full_sha1_from_ambiguous_commits_pre_2_31(git_repo):
    # 2 ambiguous refs:
    # branch ambiguous_1 - 95da0b576271cb5bee5f3e075074c03ee05fed05
    # branch ambiguous_2 - 95da0b57a416d9c8ce950554228d1fc195c30b43
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="95da0b57"
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == (
        "error: short SHA1 95da0b57 is ambiguous\n"
        "hint: The candidates are:\n"
        "hint:   95da0b576 commit 2019-10-22 - Ambiguous commit 16955\n"
        "hint:   95da0b57a commit 2019-10-22 - Ambiguous commit 44414\n"
        "fatal: Needed a single revision"
    )


@pytest.mark.skipif(
    GIT_VERSION >= (2, 31), reason="This is test for output from Git older than 2.31."
)
@pytest.mark.asyncio
async def test_git_get_full_sha1_from_ambiguous_tag_and_commit_pre_2_31(git_repo):
    # 2 ambiguous refs:
    # branch ambiguous_with_tag - c6f0e5ec04d99bdf8c6c78ff20d66d286eecb3ea
    # tag ambiguous_tag_66387 - c6f0e5ec04d99bdf8c6c78ff20d66d286eecb3ea
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_FULL_SHA1, path=git_repo.folder_path, rev="c6f0"
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == (
        "error: short SHA1 c6f0 is ambiguous\n"
        "hint: The candidates are:\n"
        "hint:   c6f028f tag ambiguous_tag_66387\n"
        "hint:   c6f0e5e commit 2019-10-24 - Commit ambiguous with tag.\n"
        "fatal: Needed a single revision"
    )


@pytest.mark.asyncio
async def test_git_is_ancestor_true(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_IS_ANCESTOR,
            path=git_repo.folder_path,
            maybe_ancestor_rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
            descendant_rev="fb99eb7d2d5bed514efc98fe6686b368f8425745",
        )
    )
    assert p.returncode == 0


@pytest.mark.asyncio
async def test_git_is_ancestor_false(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_IS_ANCESTOR,
            path=git_repo.folder_path,
            maybe_ancestor_rev="fb99eb7d2d5bed514efc98fe6686b368f8425745",
            descendant_rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
        )
    )
    assert p.returncode == 1


@pytest.mark.asyncio
async def test_git_is_ancestor_invalid_object(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_IS_ANCESTOR,
            path=git_repo.folder_path,
            maybe_ancestor_rev="invalid1",
            descendant_rev="invalid2",
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == "fatal: Not a valid object name invalid1"


@pytest.mark.asyncio
async def test_git_is_ancestor_invalid_commit(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_IS_ANCESTOR,
            path=git_repo.folder_path,
            maybe_ancestor_rev="0123456789abcde0123456789abcde0123456789",
            descendant_rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == (
        "fatal: Not a valid commit name 0123456789abcde0123456789abcde0123456789"
    )


@pytest.mark.asyncio
async def test_git_check_if_module_exists_true(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECK_IF_MODULE_EXISTS,
            path=git_repo.folder_path,
            rev="fb99eb7d2d5bed514efc98fe6686b368f8425745",
            module_name="mycog",
        )
    )
    assert p.returncode == 0


@pytest.mark.asyncio
async def test_git_check_if_module_exists_false(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_CHECK_IF_MODULE_EXISTS,
            path=git_repo.folder_path,
            rev="a7120330cc179396914e0d6af80cfa282adc124b",
            module_name="mycog",
        )
    )
    assert p.returncode == 128
    assert p.stderr.decode().strip() == (
        "fatal: Not a valid object name a7120330cc179396914e0d6af80cfa282adc124b:mycog/__init__.py"
    )


@pytest.mark.asyncio
async def test_git_find_last_occurrence_existent(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_LAST_MODULE_OCCURRENCE_COMMIT,
            path=git_repo.folder_path,
            descendant_rev="2db662c1d341b1db7d225ccc1af4019ba5228c70",
            module_name="mycog",
        )
    )
    assert p.returncode == 0
    # the command gives a commit after last occurrence
    assert p.stdout.decode().strip() == "a7120330cc179396914e0d6af80cfa282adc124b"


@pytest.mark.asyncio
async def test_git_find_last_occurrence_non_existent(git_repo):
    p = await git_repo._run(
        ProcessFormatter().format(
            git_repo.GIT_GET_LAST_MODULE_OCCURRENCE_COMMIT,
            path=git_repo.folder_path,
            descendant_rev="c950fc05a540dd76b944719c2a3302da2e2f3090",
            module_name="mycog",
        )
    )
    assert p.returncode == 0
    assert p.stdout.decode().strip() == ""
